from ._optproblems import OneMax
from ._optproblems import Sphere
from ._optproblems import Schwefe1_2
from ._optproblems import Rosenbrock
from ._optproblems import Rastrigin
from ._optproblems import HighConditionedElliptic
from ._optproblems import Griewank
from ._optproblems import Ackley
from ._optproblems import Weierstrass